#include <math.h>
#include <stdio.h>
#include <device.h>
#include <fcntl.h>
#include <stdlib.h>
#include <fmclient.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <gl/gl.h>
#include <gl/device.h>

/*
  expt_no = 0 
   rings 
  expt_no = 1 
   pinwheels
  expt_no = 2
   checkerboards
  expt_no = 3   (special titration, with rings)
*/


#define WXDIM     660
#define WYDIM     505
#define WARP      ((float)WXDIM/(float)WYDIM)
#define FIXPT     3
#define FIXCOLOR  255
#define NO_PASSES 5
#define NAPDELAY  1
#define XDIM      250
#define YDIM      250

#define ZDIM      32
#define SECS      30

#define PI        3.14159
#define RINGS     0
#define WEDGE     1
#define CHECKERBOARD   2
#define XSHIFT    (WXDIM-(XDIM*2))/2
#define YSHIFT    (WYDIM-(YDIM*2))/2
float SCALE=.001;
int NRINGS=6, NPINWHEELS=6, MAXPOINTS=3000, contrast, offset1, offset2;
int STATIC2=0;
float SPEED, DANG;


#define XSIZE     XDIM
#define YSIZE     YDIM
#define FXPT      (float)XSIZE/40.0

#define KRAD       (double)(2.0*NRINGS)/(XSIZE)
#define KANG       (double)(2*NPINWHEELS)/(2.0*PI)

int expt_no, bytes_pixel;
unsigned char stra[256], strb[256];
unsigned char ***stimulusa, ***stimulusb;
unsigned long **lstimulusa, **lstimulusb, *fixation;
unsigned char ***i1a, ***i2a, ***i3a, ***i4a;
float **ma, **disc;
int  xdim, ydim, zdim, frame_number=0;

static long start = 0;
int passes = 0, next=0, count=NAPDELAY, inc=1, running = 0, done = 0;
struct timeval tp;

#define  WHITE_LONG  255*256*256+255*256+255

generate_fixation(unsigned long* fix, int xdim, int ydim)
{
  unsigned char ch;
  int i, j, k;
  int yc = ydim/2, xc = xdim/2, dy = FIXPT, dx = FIXPT;

  for (j=yc-dy; j<yc+dy; j++) 
    for (i=xc-dx; i<xc+dx; i++) {
      *(fix+xdim*j+i) = 0;
    }
}

charseq_longseq(unsigned char*** in, unsigned long** out, int xdim, int ydim, int zdim)
{
  unsigned char ch;
  int i, j, k;

  for (k=0; k<zdim; k++) {
    for (j=0; j<ydim; j++)
      for (i=0; i<xdim; i++) {
	ch = in[k][j][i]; 
	*(out[k]+xdim*j+i) = ch*256*256+ch*256+ch;
      }
    generate_fixation(out[k], xdim, ydim); 
  }
}


assemble_i(unsigned char*** i1, unsigned char*** i2, float** m,  
	   unsigned char*** stimulus) 
{
  float fi1, fi2, fm;
  int x, y, z;
  unsigned char *i1ptr, *i2ptr;
  float *mptr;

  for (z=0; z<zdim; z++) {
    printf("."); fflush(stdout);
    for (y=0; y<ydim; y++) {
      mptr = m[y];
      i1ptr = i1[z][y];
      i2ptr = i2[z][y];
      for (x=0; x<xdim; x++) {
	fm = mptr[x];
	stimulus[z][y][x] = disc[y][x]*(unsigned char)(((fm*(float)i1ptr[x] +
							 (1.0-fm)*(float)i2ptr[x])));  
	/* stimulus[z][y][x] = 255*disc[y][x]*fm; */
      }
    }
  }
  printf("\n");
}

void add_offset(float** i1, int offset, int xdim, int ydim)
{
  int x, y;
  float f;

  for (y=0; y<ydim; y++) {
    for (x=0; x<xdim; x++) {
      f = i1[y][x] + offset;
      if ( f < 0.0 ) 
	f = 0.0;
      if ( f > 255.0 )
	f = 255.0;
      i1[y][x] = f;
    }
  }
  
}

void draw_picture()
{
  int i;

  RGBcolor(0,0,0);
  clear();
  RGBcolor(255, 255, 255);
  if ( running ) {
      sginap(count);
      frame_number += inc;
      gettimeofday(&tp);
      if ( ( next == 0 ) && ( tp.tv_sec - start >= SECS ) ) {
	next = 1;
	frame_number = 0;
	inc =1;
      }
      if ( ( next == 1 ) && ( tp.tv_sec - start >= SECS*2 ) ) {
	next = 0;
	start = tp.tv_sec;
	if ( ++passes >= NO_PASSES ) 
	  exit(0);
	frame_number = 0;
	inc =1;
      }
      if ( frame_number < 0 ) {
	inc = 1;
	frame_number = 1;
      }
      if ( frame_number > zdim-1 ) {
	inc = -1;
	frame_number = zdim-2;
      }
  }
  if ( running ) {
    rectzoom(2,2);
    if ( next ) 
      lrectwrite(0+XSHIFT, 0+YSHIFT, xdim-1+XSHIFT, ydim-1+YSHIFT, lstimulusb[frame_number]);
    else
      lrectwrite(0+XSHIFT, 0+YSHIFT, xdim-1+XSHIFT, ydim-1+YSHIFT, lstimulusa[frame_number]);
  } 
  swapbuffers();
}

void rotate_picture(float **in, unsigned char **out, float theta, int xdim, int ydim)
{
  int x, y,  ix, iy, xpi, ypi;
  int xc = xdim/2, yc = ydim/2;
  float rad, ang, fx, fy, dx, dy, xpf, ypf, ans;

  for (y=0; y<ydim; y++) {
    for (x=0; x<xdim; x++) {
      dx = x-xc;
      dy = y-yc;
      rad = sqrt((float)(dx*dx+dy*dy));
      ang = atan2((float)dy, (float)dx) + theta;
      xpf = xc+rad*cos(ang);
      ypf = yc+rad*sin(ang);
      
      xpi = (int)floor(xpf);
      ypi = (int)floor(ypf);
      if ( ( xpi < 0 ) || ( ypi < 0 ) || ( xpi >= xdim-1) ||  ( ypi >= ydim-1) ) 
        out[y][x] = 0;
      else {
        dx  = xpf-(float)xpi;
        dy  = ypf-(float)ypi;
	ans = 
	  (1-dx) * (1-dy) * in[ypi][xpi] + 
	  dx   * (1-dy) * in[ypi][xpi+1] +
	  (1-dx) *   dy   * in[ypi+1][xpi] +
	  dx   *   dy   * in[ypi+1][xpi];
	if ( ans > 255 ) 
	  ans = 255.0;
        out[y][x] = (unsigned char)ans;
      }
    }
  }
}

void translate_picture(float **in, unsigned char **out, 
		       float xshift, float yshift, int xdim, int ydim)
{
  int x, y,  ix, iy, xpi, ypi;
  int xc = xdim/2, yc = ydim/2;
  float rad, ang, fx, fy, dx, dy, xpf, ypf;

  for (y=0; y<ydim; y++) {
    for (x=0; x<xdim; x++) {
      xpi = (x + (int)xshift + xdim)%xdim;
      ypi = (y + (int)yshift + ydim)%ydim;
      out[y][x] = (unsigned char)in[ypi][xpi];
    }
  }
}

#define K2   .8

void generate_mask(float** mask, int xdim, int ydim)
{
  int xc = xdim/2, yc = ydim/2;
  float rad, ang, qr;
  float kang = KANG, krad = KRAD;
  int x, y, dx, dy, q, points=0, qx, qy;

  if ( expt_no != CHECKERBOARD ) {
    for (y=0; y<ydim; y++) {
      for (x=0; x<xdim; x++) {
	dx = x-xc;
	dy = y-yc;
	rad = sqrt((float)(dx*dx+dy*dy));
	ang = atan2((float)dy, (float)dx);
	switch ( expt_no ) {
	case WEDGE: 
	  q = (int)(kang*ang+1000); 
	  if ( q % 2 == 0 ) {
	    mask[y][x] = 1.0;
	    points++;
	  } else
	    mask[y][x] = 0;
	  break;
	case 3:
	case RINGS:
	  qr = rad/XSIZE;
	  krad = 4.0*PI*NRINGS*(1.0 - K2*qr);
	  if ( sin(krad*qr) > 0.0 )
	    mask[y][x] = 1.0;
	  else 
	    mask[y][x] = 0.0;
	  break;
	default:
	  mask[y][x] = 1.0;
	}
      }
    }
  } else {
    for (y=0; y<ydim; y++) {
      for (x=0; x<xdim; x++) {
	dx = KRAD*x;
	dy = KRAD*y;
	qx = ((int)dx+1000)%2;
	qy = ((int)dy+1000)%2;
	if ( ( ( qx == 0 ) && ( qy == 0 ) ) || ( (qx && qy) ) )
	  mask[y][x] = 1.0;
	else 
	  mask[y][x] = 0.0;
      } 
    }
  }
}

void generate_disc(float** mask, int xdim, int ydim)
{
  int xc = xdim/2, yc = ydim/2;
  float rad, ang;
  float kang = KANG, krad = KRAD;
  int x, y, dx, dy, q, points=0;

  if ( expt_no != CHECKERBOARD ) {
    for (y=0; y<ydim; y++) {
      for (x=0; x<xdim; x++) {
	dx = x-xc;
	dy = y-yc;
	rad = sqrt((float)(dx*dx+dy*dy));
	if ( rad < XSIZE/2 ) 
	  mask[y][x] = 1.0;
	else
	  mask[y][x] = 0.0;
      }
    }
  } else {
    for (y=0; y<ydim; y++) 
      for (x=0; x<xdim; x++) 
	  mask[y][x] = 1.0;

  }

}

void make_stuff()
{
  int bytes_pixel, x, y, z;
  float gauss[3] = { .25, .5, .25 };
  float gauss1[3] = { .1, .8, .1 };
  float** noise1_unblurred;
  float** noise2_unblurred;
  float** noise3_unblurred;
  float** noise4_unblurred;
  float** noise1;
  float** noise2;
  float** noise3;
  float** noise4;
  float** tmp;
  float** tmp2;
  float theta, shift;

  xdim = XDIM;
  ydim = YDIM;
  zdim = ZDIM;

  printf("Generating stimulus [1]\n");
  tmp = (float**)allocate_picture(xdim, ydim, sizeof(float));
  tmp2 = (float**)allocate_picture(xdim, ydim, sizeof(float));
  ma = (float**)allocate_picture(xdim, ydim,  sizeof(float));
  disc = (float**)allocate_picture(xdim, ydim,  sizeof(float));

  generate_mask(tmp, xdim, ydim);

  apply_separable_2d_filter(tmp, ma, tmp2, xdim, ydim,
			    gauss1, gauss1, 3, 3, 1, 1);
  generate_disc(tmp, xdim, ydim);
  apply_separable_2d_filter(tmp, disc, tmp2, xdim, ydim,
			    gauss1, gauss1, 3, 3, 1, 1);

  noise1_unblurred = (float**)allocate_picture(xdim, ydim, sizeof(float));
  noise2_unblurred = (float**)allocate_picture(xdim, ydim, sizeof(float));
  noise1 = (float**)allocate_picture(xdim, ydim, sizeof(float));
  noise2 = (float**)allocate_picture(xdim, ydim, sizeof(float));

  for (y=0; y<ydim; y++) {
    for (x=0; x<xdim; x++) {
      noise1_unblurred[y][x] = (float)(rand()%contrast);
      noise2_unblurred[y][x] = (float)(rand()%contrast);
    }
  }

  apply_separable_2d_filter(noise1_unblurred, noise1, tmp, xdim, ydim, gauss, gauss, 3, 3, 1, 1);
  apply_separable_2d_filter(noise2_unblurred, noise2, tmp, xdim, ydim, gauss, gauss, 3, 3, 1, 1);

  i1a = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));
  i2a = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));
  printf("Generating stimulus [2]\n");
  if ( expt_no == CHECKERBOARD ) {
    for (z=0; z<zdim; z++) {
      printf(".");   fflush(stdout);
      shift = 20.0*SPEED*((float)z)/((float)zdim);
      if ( STATIC2 ) 
	translate_picture(noise1, i1a[z], 0.0, 0.0, xdim, ydim);
      else 
	translate_picture(noise1, i1a[z], shift, 0.0, xdim, ydim);
      translate_picture(noise2, i2a[z], -1.0*shift, 0.0, xdim, ydim);
    }
  } else {
    for (z=0; z<zdim; z++) {
      printf(".");   fflush(stdout);
      theta = SPEED*(PI*(float)z)/(4.0*(float)zdim);
      if ( STATIC2 ) 
	rotate_picture(noise1, i1a[z], 0.0, xdim, ydim);
      else 
	rotate_picture(noise1, i1a[z], theta, xdim, ydim);
      rotate_picture(noise2, i2a[z], -1.0*theta, xdim, ydim);
    }
  }

  printf("Generating stimulus [3]\n");
  stimulusa = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));
  stimulusb = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));

  fixation = (unsigned long*)malloc(sizeof(unsigned long)*xdim*ydim);
  generate_fixation(fixation, xdim, ydim);
  assemble_i(i1a, i2a, ma, stimulusa);
  
  if ( expt_no == 3 ) {
    printf("Doing EXPERIMENT 3\n");
    noise3 =(float**)allocate_picture(xdim, ydim, sizeof(float));
    noise3_unblurred = (float**)allocate_picture(xdim, ydim, sizeof(float));
    noise4 =(float**)allocate_picture(xdim, ydim, sizeof(float));
    noise4_unblurred = (float**)allocate_picture(xdim, ydim, sizeof(float));
    for (y=0; y<ydim; y++) {
      for (x=0; x<xdim; x++) {
	noise3_unblurred[y][x] = (float)((rand()%contrast));
	noise4_unblurred[y][x] = (float)((rand()%contrast));
      }
    }
    apply_separable_2d_filter(noise3_unblurred, noise3, tmp, xdim, ydim, gauss, gauss, 3, 3, 1, 1);
    apply_separable_2d_filter(noise4_unblurred, noise4, tmp, xdim, ydim, gauss, gauss, 3, 3, 1, 1);
    add_offset(noise3, offset1, xdim, ydim);
    add_offset(noise4, -offset2, xdim, ydim);

    i3a = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));
    i4a = (unsigned char***)allocate_sequence(xdim, ydim, zdim, sizeof(unsigned char));
    for (z=0; z<zdim; z++) {
      printf(".");   fflush(stdout);
      theta = SPEED*(PI*(float)z)/(4.0*(float)zdim);
      rotate_picture(noise3, i3a[z], theta, xdim, ydim);
      rotate_picture(noise4, i4a[z], theta, xdim, ydim);
    }
    assemble_i(i3a, i4a, ma, stimulusb);
  } else 
    assemble_i(i2a, i1a, disc, stimulusb);

  printf("Generating stimulus [4]\n");
  write_descriptor(stra, xdim, ydim, zdim, sizeof(unsigned char));
  write_sequence(stra, stimulusa, xdim, ydim, zdim, sizeof(unsigned char));
  write_descriptor(strb, xdim, ydim, zdim, sizeof(unsigned char));
  write_sequence(strb, stimulusb, xdim, ydim, zdim, sizeof(unsigned char));
} 

draw_fxpt(int col)
{
  double vec[3];
  RGBcolor(col, col, col);
  bgnpolygon();
  vec[2] = 0.0;
  vec[0] = -FXPT;
  vec[1] = -FXPT;
  v2d(vec);
  vec[0] = FXPT;
  vec[1] = -FXPT;
  v2d(vec);
  vec[0] = FXPT;
  vec[1] = FXPT;
  v2d(vec);
  vec[0] = -FXPT;
  vec[1] = FXPT;
  v2d(vec);
  vec[0] = -FXPT;
  vec[1] = -FXPT;
  v2d(vec);
  endpolygon();
}

main(int argc, char *argv[])
{
  FILE *Infile;
  int x, y, z, k, magic, jnk;
  short val, done, mval[2], lastval[2];
  Device dev, mdev[2];
  struct stat stbuf;
  char ch;

  printf("LUMNOISE - fMRI Motion boundary stimuli (c) 1996  By Niyogi/Reppas\n");
  printf("Usage: lumnoise [exptno] [rings/pinwheels] [STATIC2] [speed] [offset1] [offset2]\n");
  printf(" [exptno: 0=rings  1=pinwheels, 2=checkerboard 3=CONTRAST -> must be 3]\n");
  printf(" [STATIC2: should the second surface be static (0/1)\n");
  printf(" [rings/pinwheels: # of rings/pinwheels (depends on exptno, def 6)]\n");
  printf(" [speed: modulation factor (.1 to 10.0, def 1.0)]\n");
  printf(" [offset1: + shift offset of light strip (0 to 255, def 0)]\n");
  printf(" [offset2: -                 dark strip (0 to 255, def 0)]\n");

  if ( argc < 2 ) 
    exit(0);

  ch = argv[1][0];
  if ( ( ( ch >= '1' ) && ( ch <= '9') ) || ( ch == '0' ) ) {
    expt_no = atoi(argv[1]);
    if ( argc > 2 ) {
      NRINGS = atoi(argv[2]);
      NPINWHEELS = atoi(argv[2]);
    }
    
    if ( argc > 3 ) 
      STATIC2 = (double)atoi(argv[3]);
    else
      STATIC2 = 0;
    
    if ( argc > 4 ) 
      SPEED = (double)atof(argv[4]);
    else
      SPEED = 1.0;
    
    offset1 = 0;
    offset2 = 0;

    if ( argc > 5 )
      offset1 = atoi(argv[5]);
    if ( argc > 6 )
      offset2 = atoi(argv[6]);
    contrast = 255;

    if ( ( expt_no == 0 ) || ( expt_no == 3 ) )
      k = NRINGS;
    else
      k = NPINWHEELS;
    sprintf(stra, "stima-%d-%d-%d-%2.2f-%d-%d", expt_no, k, STATIC2, SPEED, offset1, offset2);
    sprintf(strb, "stimb-%d-%d-%d-%2.2f-%d-%d", expt_no, k, STATIC2, SPEED, offset1, offset2);
  } else {
    sprintf(stra, "%s", argv[1]);
    sprintf(strb, "%s", argv[2]);
  }
  if ( (stat(stra, &stbuf) == -1) || (stat(strb, &stbuf) == -1) ) {
    make_stuff();
  } else {
    read_descriptor(stra, &xdim, &ydim, &zdim, &bytes_pixel);
    read_descriptor(strb, &xdim, &ydim, &zdim, &bytes_pixel);
    stimulusa = (unsigned char***)allocate_sequence(xdim, ydim, zdim,
						    sizeof(unsigned char));
    stimulusb = (unsigned char***)allocate_sequence(xdim, ydim, zdim,
						    sizeof(unsigned char));
    read_sequence(stra, stimulusa, xdim, ydim, zdim, sizeof(char));
    read_sequence(strb, stimulusb, xdim, ydim, zdim, sizeof(char));
  }

  lstimulusa = (unsigned long**)malloc(sizeof(unsigned long*)*zdim);
  for (z=0; z<zdim; z++) 
    lstimulusa[z] = (unsigned long*)malloc(sizeof(unsigned long)*xdim*ydim);
  lstimulusb = (unsigned long**)malloc(sizeof(unsigned long*)*zdim);
  for (z=0; z<zdim; z++) 
    lstimulusb[z] = (unsigned long*)malloc(sizeof(unsigned long)*xdim*ydim);
  charseq_longseq(stimulusa, lstimulusa, xdim, ydim, zdim);
  charseq_longseq(stimulusb, lstimulusb, xdim, ydim, zdim);

  /* Initialize graphics and build window */
  prefposition(144,802,472,976);
  prefsize(WXDIM, WYDIM);

  foreground();
  winopen("bluehex");
  RGBmode();
  doublebuffer();
  gconfig();
  subpixel(TRUE);
  linesmooth(SML_ON);
 
  ortho2(-XSIZE*WARP, XSIZE*WARP, -YSIZE, YSIZE);

  qdevice(LEFTMOUSE);
  qdevice(RIGHTMOUSE);
  qdevice(ESCKEY);
  qdevice(F1KEY);
  qdevice(F2KEY);

  RGBcolor(0,0,0);
  clear();
  draw_fxpt(255);
  swapbuffers();

  printf("Type a number\n");
  scanf("%d", &jnk);

  running = 1;
  done = 0;
  gettimeofday(&tp);
  start = tp.tv_sec;
  while ( ! done ) {
    if ( running ) 
      draw_picture();
    if (qtest()) {
      switch (dev = qread(&val)) {
      case F1KEY:	      
	running=1;
	gettimeofday(&tp);
	start = tp.tv_sec;
	break;
      case F2KEY:	    
	running=0;
	break;
      case ESCKEY:
 	done = 1;
	exit(0);  
	break;
      }
    }
  }
  return(0);
}










