#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <float.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <math.h>
#include "tool.h"

double fabs(double f) { if ( f >= 0.0 ) return(f); else return(-f); }

int pow2(int n)
{
  int ans=1;
  while ( n-- > 0 ) 
    ans *= 2;
  return(ans);
}

void draw_line(unsigned char **pic, float p0[2], float p1[2], int xdim, int ydim, unsigned char val)
{
  float t, tx, ty;
  int ix, iy;

  for (t=0.0; t<=1.0; t+= .01) {
    tx = (p1[0]-p0[0])*t + p0[0];
    ty = (p1[1]-p0[1])*t + p0[1];
    ix = (xdim+(int)tx)%xdim;   
    iy = (ydim+(int)ty)%ydim;
    pic[iy][ix] = val;
  }
}

int random_integer(int s, int e)
{
  return((rand()%(e-s))+s);
}

float random_float(float f0, float f1)
{
  int r = rand()%10000;
  float fr = (float)r/10000.0;

  return( f0 + (f1-f0)*fr );
}


float energy_image(float **image, int xdim, int ydim)
{
  int y, x;
  float val;

  val = 0.0;
  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      val += image[y][x]*image[y][x];
  return(val);
}

float energy_vector(float *vec, int dim)
{
  int  x;
  float val;

  val = 0.0;
  for (x=0; x<dim; x++)
      val += vec[x]*vec[x];
  return(val);
}


float weighted_euclidean_distance_vector(float *v1, float *v2, float *w, int dim)
{
  int  x;
  float val, tmp;

  val = 0.0;
  for (x=0; x<dim; x++) {
    tmp = (v1[x] - v2[x]);
    val += w[x]*tmp*tmp;
  }
  return(val);
}

float euclidean_distance_vector(float *v1, float *v2, int dim)
{
  int  x;
  float val, tmp;

  val = 0.0;
  for (x=0; x<dim; x++) {
    tmp = v1[x] - v2[x];
    val += tmp*tmp;
  }
  return(val);
}

float mahalanobis_distance_vector(float *v1, float *v2, float *eigenval, int dim)
{
  int  x;
  float val, tmp;

  val = 0.0;
  for (x=0; x<dim; x++) {
    tmp = v1[x] - v2[x];
    val += (tmp*tmp)/eigenval[x];
  }
  return(val);
}

void copy_picture(float **im1, float **im2, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      im2[y][x] = im1[y][x];
}


void add_image(float **im1, float **im2, float **res, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      res[y][x] = im1[y][x] + im2[y][x];
}

void subtract_image(float **im1, float **im2, float **res, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      res[y][x] = im1[y][x] - im2[y][x];
}

void add_cimage(float **im1, float c, float **res, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      res[y][x] = im1[y][x] + c;
}

void divide_image(float **im1, float **im2, float **res, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      res[y][x] = im1[y][x] / im2[y][x];
}

void fill_image(float **res, float val, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++)
    for (x=0; x<xdim; x++)
      res[y][x] = val;
}
/*
void make_xtslice(in, out, y_loc, xdim, ydim, nframe)
unsigned char *in[MAXFRAMES];
unsigned char *out;
int y_loc, xdim, ydim, nframe;
{
  int x, n;
  for (x=0; x<xdim; x++)
    for (n=0; n<nframe; n++)
      *(out+n*xdim+x) = *(in[n]+y_loc*xdim+x);
}
*/
void **allocate_picture(int xdim, int ydim, int size)
{
  int j;
  void **seq = (void**)malloc(sizeof(void*)*ydim);
  if ( ! seq ) { printf("out of memory\n"); exit(0); }
  for (j=0; j<ydim; j++) {
    seq[j] = (void*)malloc(size*xdim);
    if ( ! seq[j] ) { printf("out of memory\n"); exit(0); }
  }
  return(seq);
} 

void free_picture(void **pic, int xdim, int ydim)
{
  int j;

  for (j=0; j<ydim; j++) 
    free(pic[j]);
  free(pic);
} 

void ***allocate_sequence(int xdim, int ydim, int zdim, int size)
{
  int j, k;
  void ***seq = (void***)malloc(sizeof(void**)*zdim);
  if ( ! seq ) { printf("out of memory\n"); exit(0); }
  for (k=0; k<zdim; k++) {
    seq[k] = (void**)malloc(sizeof(void**)*ydim);
    if ( ! seq[k] ) { printf("out of memory\n"); exit(0); }
    for (j=0; j<ydim; j++) {
      seq[k][j] = (void*)malloc(size*xdim);
      if ( ! seq[k][j] ) { printf("out of memory\n"); exit(0); }
    }
  }
  return(seq);
} 

void free_sequence(void ***seq, int xdim, int ydim, int zdim)
{
  int j, k;

  for (k=0; k<zdim; k++) {
    for (j=0; j<ydim; j++) 
      free(seq[k][j]);
    free(seq[k]);
  }
  free(seq);
} 

void sum_sequence(float ***seq1, float ***seq2, float ***out, int xdim, int ydim, int zdim)
{
  int x, y, z;

  for (z=0; z<zdim; z++)
    for (y=0; y<ydim; y++)
      for (x=0; x<xdim; x++)
	out[z][y][x] = seq1[z][y][x]+seq2[z][y][x];
}

void fill_sequence(float ***seq, float val, int xdim, int ydim, int zdim)
{
  int x, y, z;

  for (z=0; z<zdim; z++)
    for (y=0; y<ydim; y++)
      for (x=0; x<xdim; x++)
	seq[z][y][x] = val;
}

void square_sequence(float ***seq, float*** out, int xdim, int ydim, int zdim)
{
  int x, y, z;
  float tmp;

  for (z=0; z<zdim; z++)
    for (y=0; y<ydim; y++)
      for (x=0; x<xdim; x++) {
	tmp = seq[z][y][x];
	out[z][y][x] = tmp*tmp;
      }
}


float maximum_sequence(float ***seq, int xdim, int ydim, int zdim)
{
  int x, y, z;
  float max = 0.0, tmp;

  for (z=0; z<zdim; z++)
    for (y=0; y<ydim; y++)
      for (x=0; x<xdim; x++) {
	tmp = fabs(seq[z][y][x]);
	if ( tmp > max )
	  max = tmp;
      }
  return(max);
}

float maximum_picture(float **seq, int xdim, int ydim)
{
  int x, y;
  float max = 0.0, tmp;

    for (y=0; y<ydim; y++)
      for (x=0; x<xdim; x++) {
	tmp = fabs(seq[y][x]);
	if ( tmp > max )
	  max = tmp;
      }
  return(max);
}

float maximum_vector(float *in, int dim)
{
  int x;
  float max = 0.0, tmp;

  for (x=0; x<dim; x++) {
    tmp = fabs(in[x]);
    if ( tmp > max )
      max = tmp;
  }
  return(max);
}


void convert_sequence_char_float(unsigned char ***in, float ***out, 
				 int xdim, int ydim, int zdim)
{
  int i, j, k;
  float scale = 1.0/255.0;

  for (k=0; k<zdim; k++)
    for (j=0; j<ydim; j++) 
      for (i=0; i<xdim; i++) 
	out[k][j][i] = (float)in[k][j][i]*scale;
}

void convert_sequence_float_char(float*** in, unsigned char ***out, 
				 int xdim, int ydim, int zdim, int method)
{
  int i, j, k, tmp;
  float max = maximum_sequence(in, xdim, ydim, zdim);
  float scale;

  switch ( method ) {
  case SAVE_POSNEG: 
    scale = 128.0/max; 
    for (k=0; k<zdim; k++)
      for (j=0; j<ydim; j++) 
	for (i=0; i<xdim; i++) {
	  tmp = (int)((in[k][j][i]*scale)+128.0);
	  if ( tmp < 0 )
	    tmp = 0;
	  if ( tmp > 255 )
	    tmp = 255;
	  out[k][j][i] = (unsigned char)tmp;
	}
    break;
  case SAVE_POS:
    scale = 255.0/max;
    for (k=0; k<zdim; k++)
      for (j=0; j<ydim; j++)
	for (i=0; i<xdim; i++) {
	  tmp = (int)((in[k][j][i]*scale));
	  if ( tmp < 0 )
	    tmp = 0;
	  if ( tmp > 255 )
	    tmp = 255;
	  out[k][j][i] = (unsigned char)tmp;
	}
    break;
  default: printf("convert_sequence_float_char: Unknown format.\n"); break;
  }
}

void convert_char_float(unsigned char **in, float** out, 
			int xdim, int ydim, int method)
{
  float scale = 1.0/255.0;
  int i, j;

  if ( method == 1 ) 
    scale = 1.0/255.0;
  else
    scale = 1.0;

  for (j=0; j<ydim; j++)
    for (i=0; i<xdim; i++)
      out[j][i] = (float)in[j][i]*scale;
}

void convert_float_char_vector(float *in, unsigned char *out, int dim)
{
  float max = maximum_vector(in, dim);
  int i;

  max = 255.0/max;
  for (i=0; i<dim; i++)
    out[i] = (unsigned char)in[i]*max;
}

void convert_float_char(float** in, unsigned char** out, 
			int xdim, int ydim, int method)
{
  int i, j, tmp;
  float max = maximum_picture(in, xdim, ydim);
  float scale;

  switch ( method ) {
  case SAVE_POSNEG: 
    if ( fabs(max) < .0001 )  {
      printf("oops.\n");
      scale = 1.0;
    } else 
      scale = 128.0/max; 
    printf("scale %f\n", scale);
    for (j=0; j<ydim; j++) 
	for (i=0; i<xdim; i++) {
	  tmp = (int)((in[j][i]*scale)+128.0);
	  if ( tmp < 0 )
	    tmp = 0;
	  if ( tmp > 255 )
	    tmp = 255;
	  out[j][i] = (unsigned char)tmp;
	}
    break;
  case SAVE_POS:
    if ( fabs(max) < EPS )  {
      printf("oops: %f\n", max);
      scale = 1.0;
    } else 
      scale = 255.0/max;
    for (j=0; j<ydim; j++)
      for (i=0; i<xdim; i++) {
	tmp = (int)((in[j][i]*scale));
	if ( tmp < 0 )
	  tmp = 0;
	if ( tmp > 255 )
	  tmp = 255;
	out[j][i] = (unsigned char)tmp;
      }
    break;
  default: printf("convert_sequence_float_char: Unknown format.\n"); break;
  }
}

void read_coeffs(char* indir, char* tname, float **coeff,
	    int *nimages, int *ncoeffs)
{
  FILE *fp1, *fp2;
  int done, c, d;
  char filename[STRING], str[STRING];

  fp1 = fopen(tname, "r");
  if ( ! fp1 ) {
    printf("Couldn't open [%s]\n", tname);
    exit(0);
  }
  for (done=0, c=0; ! feof(fp1); c++) {
    fscanf(fp1, "%s", str);
    sprintf(filename, "%s/%s", indir, str);
    fp2 = fopen(filename, "r");
    if ( ! fp2 )
      done = 1;
    else {
      for (d=0; ! feof(fp2); d++) {
	fscanf(fp2, "%s", str);
	coeff[c][d] = atof(str);
      }
      *ncoeffs = d-1;
    }
    fclose(fp2);
  }
  fclose(fp1);
  *nimages = c-1;
  printf("Read %d coeffs %d images\n", *ncoeffs, *nimages);
}

int read_ascii_float1d(char* file, float* arr, int* dim, int max_dim)
{
  FILE* fp;
  struct stat stbuf;
  int d;
  char str[STRING];

  *dim  =0;
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    return(0);
  }    
  if ((fp = fopen(file, "r")) == NULL) {
    printf("Cannot open input file %s\n", file);
    return(0);
  }
  for (d=0; ! feof(fp) && d < max_dim; d++) {
    fscanf(fp, "%s", str);
    arr[d] = atof(str);
  }
  *dim = d-1;
  return(1);
}

void read_ascii_float2d(char* file, char* indir, 
			float** arr, int* xdim, int* ydim, 
			int max_xdim, int max_ydim)
{
  FILE* fp;
  struct stat stbuf;
  int c, d;
  char fname[STRING], str[STRING];

  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }
  if ((fp = fopen(file, "r")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  *xdim = 0;
  for (d=0; ! feof(fp) && d < max_ydim; d++) {
    fscanf(fp, "%s", str);
    sprintf(fname, "%s/%s", indir, str);
    printf("Reading %s\n", fname);
    if ( read_ascii_float1d(fname, arr[d], &c, max_xdim) )
      if ( c > *xdim )
	*xdim = c;
  }
  *ydim = d-1;
}


void read_picture(char* name, void **pic, int xdim, int ydim, int bytes_pixel)
{
  char file[STRING];
  FILE* f;
  struct stat stbuf;
  int y;

  sprintf(file, "%s/data", name);
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }    
  if ((f = fopen(file, "rb")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  for (y=0; y<ydim; y++) 
    fread(pic[y], bytes_pixel, xdim, f); 
  fclose(f);
}

void read_vector(char* name, void* pic, int xdim, int bytes_pixel)
{
  char file[STRING];
  FILE* f;
  struct stat stbuf;


  sprintf(file, "%s/data", name);
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }    
  if ((f = fopen(file, "rb")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  fread(pic, bytes_pixel, xdim, f); 
  fclose(f);
}

void read_matrix(char* name, float** mat, int xdim, int ydim)
{
  char file[STRING];
  FILE* f;
  struct stat stbuf;
  int y;

  sprintf(file, "%s/data", name);
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }    
  if ((f = fopen(file, "rb")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  for (y=1; y<=ydim; y++) 
    fread(mat[y]+1, sizeof(float), xdim, f);
  fclose(f);
}

void flipy_picture(unsigned char* in, unsigned char* out, int xdim, int ydim)
{
  int x, y;

  for (y=0; y<ydim; y++) {
    for (x=0; x<xdim; x++) {
      *(out+y*xdim+x) = *(in+(ydim-y-1)*xdim+x);
    }
  }
}

void read_sequence(char* name, void ***seq, int xdim, int ydim, int
		   nframe, int bytes_pixel)
{
  char file[255];
  int frame, y;
  struct stat stbuf;
  FILE* f;

  for (frame = 0; frame < nframe; frame++) {
    printf(".");  
    fflush(stdout);
    sprintf(file, "%s/data%d", name, frame);
    if ( stat(file, &stbuf) == -1 ) {
      printf("Error: Couldn't access %s. \n", file);
      exit(1);
    }    
    if ((f = fopen(file, "r")) == NULL) {
      printf("Cannot open input file %s\n", file);
      exit(1);
    }
    for (y=0; y<ydim; y++) 
      fread(seq[frame][y], bytes_pixel, xdim, f);
    fclose(f);
  }
  printf("\n");
}

void apply_separable_2d_filter(float **image, float **result, float **temp1,
			       int xdim, int ydim,
			       float* xfilt, float* yfilt, 
			       int xfdim, int yfdim, 
			       int xfilt_space, int yfilt_space)
{
  register double sum;
  int y_fmid = yfdim/2;
  int x_fmid = xfdim/2;
  int x,y,x2,y2;

  int ydim2 = (int) ((float)(ydim)/yfilt_space+.5);
  register int x_pos, y_pos, xf, yf;
  register int yoff, xoff;


  /* convolve y */
  for(y=0, y2=0; y < ydim; y+=yfilt_space,y2++) {
    yoff = y - y_fmid;
    for(x=0; x < xdim; x++) {
      sum=0.0;
      for(yf=0; yf < yfdim; yf++) {
	y_pos = yoff + yf;
	if(y_pos < 0)
	  y_pos = -y_pos;
	else
	  if(y_pos >= ydim)
	    y_pos = ydim - (y_pos - ydim + 2); 
	sum += image[y_pos][x]*yfilt[yf];
      }    
      temp1[y2][x] = sum;
    }
  }
  
  /* convolve x */
  for (x=0, x2=0; x<xdim; x+=xfilt_space, x2++) {
    xoff = x - x_fmid;
    for (y=0; y<ydim2; y++) {
      sum=0.0;
      for (xf=0; xf < xfdim; xf++) {
	x_pos = xoff + xf;
	if(x_pos < 0)
	  x_pos = -x_pos;
	else
	  if (x_pos >= xdim)
	    x_pos = xdim - (x_pos - xdim + 2); 
	sum += temp1[y][x_pos]*xfilt[xf];
      } 
      result[y][x2] = sum;
    }
  } 
}

void apply_separable_3d_filter(float ***image, float ***result, 
			       float ***temp1, float ***temp2, 
			       int xdim, int ydim, int zdim,
			       float* xfilt, float* yfilt, float* zfilt, 
			       int xfdim, int yfdim, int zfdim,
		       int xfilt_space, int yfilt_space, int zfilt_space)
{
  register double sum;
  int z_fmid = zfdim/2;
  int y_fmid = yfdim/2;
  int x_fmid = xfdim/2;
  int x,y,z,x2,y2,z2;
  int xdim2 = (int) ((float)(xdim)/xfilt_space+.5);
  int ydim2 = (int) ((float)(ydim)/yfilt_space+.5);

  register int x_pos, y_pos, z_pos, xf, yf, zf;
  register int zoff, yoff, xoff;


  printf("|-");   fflush(stdout);
  /* convolve y */
  for(y=0, y2=0; y < ydim; y+=yfilt_space,y2++) {
    yoff = y - y_fmid;
    for(z=0; z < zdim; z++) {
      for(x=0; x < xdim; x++) {
	sum=0.0;
	for(yf=0; yf < yfdim; yf++) {
	  y_pos = yoff + yf;
	  if(y_pos < 0)
	    y_pos = -y_pos;
	  else
	    if(y_pos >= ydim)
	      y_pos = ydim - (y_pos - ydim + 2); 
	  sum += image[z][y_pos][x]*yfilt[yf];
	}    
	temp1[z][y2][x] = sum;
      }
    }
  }

  printf("|+");    fflush(stdout);
  /* convolve x */
  for (x=0, x2=0; x<xdim; x+=xfilt_space, x2++) {
    xoff = x - x_fmid;
    for (y=0; y<ydim2; y++) {
      for (z=0; z<zdim; z++) {
	sum=0.0;
	for (xf=0; xf < xfdim; xf++) {
	  x_pos = xoff + xf;
	  if(x_pos < 0)
	    x_pos = -x_pos;
	  else
	    if (x_pos >= xdim)
	      x_pos = xdim - (x_pos - xdim + 2); 
	  sum += temp1[z][y][x_pos]*xfilt[xf];
	} 
	temp2[z][y][x2] = sum;
      }
    } 
  } 

  printf("|=");    fflush(stdout);
  /* convolve z */
  for (z=0, z2=0; z<zdim; z+=zfilt_space, z2++) {
    zoff = z - z_fmid;
    for (y=0; y<ydim2; y++) {
      for (x=0; x<xdim2; x++) {
	sum=0.0;
	for (zf=0; zf < zfdim; zf++) {
	  z_pos = zoff + zf;
	  if(z_pos < 0)
	    z_pos = -z_pos;
	  else
	    if (z_pos >= zdim)
	      z_pos = zdim - (z_pos - zdim + 2); 
	  sum += temp2[z_pos][y][x]*zfilt[zf];
	}
	result[z2][y][x] = sum;
      }
    }
  }
  printf("|\n");  
}

void blur_3d(float ***i, float ***o,  float ***tmp1, float ***tmp2, 
	     int xdim, int ydim, int nframe)
{
  static float gspace[5] = { 3.342604e-2, 0.241125, 0.450898, 0.241125, 3.342604e-2 };
  static float gtime[3] = { .25, .5, .25 };

  apply_separable_3d_filter(i, o, tmp1, tmp2, xdim, ydim, nframe, 
			    gspace, gspace, gtime, 5, 5, 3, 1, 1, 1);

}

float ****allocate_sequence_4d(int xdim, int ydim,int zdim, int basis_size)
{
  float ****basis = (float****)malloc(sizeof(void***)*basis_size);
  int b, j, k;

  if ( ! basis ) { printf("out of memory\n"); exit(0); }
  for (b=0; b<basis_size; b++) {
    basis[b] = (float***)malloc(sizeof(void**)*zdim);
    if ( ! basis[b] ) { printf("out of memory\n"); exit(0); }
    for (k=0; k<zdim; k++) {
      basis[b][k] = (float**)malloc(sizeof(void**)*ydim);
      if ( ! basis[b][k] ) { printf("out of memory\n"); exit(0); }
      for (j=0; j<ydim; j++) {
	basis[b][k][j] = (float*)malloc(sizeof(float)*xdim);
	if ( ! basis[b][k][j] ) { printf("out of memory\n"); exit(0); }
      }
    }
  }
  return(basis);
}

void free_sequence_4d(float ****basis, int xdim, int ydim,int zdim, int basis_size)
{
  int b, j, k;

  for (b=0; b<basis_size; b++) {
    for (k=0; k<zdim; k++) {
      for (j=0; j<ydim; j++) 
	free(basis[b][k][j]);
      free(basis[b][k]);
    }
    free(basis[b]);
  }
  free(basis);
}


void generate_3d_basis(float ***seq, float ****basis, float ***tmp1, float ***tmp2, int xdim, int ydim, int zdim, int basis_type)
{
  static float gd0[5] = { 3.342604e-2, 0.241125, 0.450898, 0.241125, 3.342604e-2 };
  static float gd1[5] = { -9.452048e-2, -0.306497, 0.00000, 0.306497, 9.452048e-2 };
  static float gd2[5] = { 0.206192, 6.440215e-2, -0.541189, 6.440215e-2, 0.206192 };
  static float gd3[5] = { -0.290333, 0.703260, 0.00000, -0.703260, 0.290333 };

  printf("Generating basis:%dx%dx%d\n", xdim, ydim, zdim);
  switch ( basis_type ) {
  case G1_BASIS:
      apply_separable_3d_filter(seq, basis[0], tmp1, tmp2, xdim, ydim, zdim, gd0, gd0, gd1, 5, 5, 5, 1, 1, 1); /* It */
      apply_separable_3d_filter(seq, basis[1], tmp1, tmp2, xdim, ydim, zdim, gd0, gd1, gd0, 5, 5, 5, 1, 1, 1); /* Iy */
      apply_separable_3d_filter(seq, basis[2], tmp1, tmp2, xdim, ydim, zdim, gd1, gd0, gd0, 5, 5, 5, 1, 1, 1); /* Ix */
    break;
  case G2_BASIS:
      apply_separable_3d_filter(seq, basis[0], tmp1, tmp2, xdim, ydim, zdim, gd0, gd0, gd2, 5, 5, 5, 1, 1, 1); /* Itt */
      apply_separable_3d_filter(seq, basis[1], tmp1, tmp2, xdim, ydim, zdim, gd0, gd1, gd1, 5, 5, 5, 1, 1, 1); /* Iyt */
      apply_separable_3d_filter(seq, basis[2], tmp1, tmp2, xdim, ydim, zdim, gd0, gd2, gd0, 5, 5, 5, 1, 1, 1); /* Iyy */
      apply_separable_3d_filter(seq, basis[3], tmp1, tmp2, xdim, ydim, zdim, gd1, gd0, gd1, 5, 5, 5, 1, 1, 1); /* Ixt */
      apply_separable_3d_filter(seq, basis[4], tmp1, tmp2, xdim, ydim, zdim, gd1, gd1, gd0, 5, 5, 5, 1, 1, 1); /* Ixy */
      apply_separable_3d_filter(seq, basis[5], tmp1, tmp2, xdim, ydim, zdim, gd2, gd0, gd0, 5, 5, 5, 1, 1, 1); /* Ixx */
    break;
  case G3_BASIS:
      apply_separable_3d_filter(seq, basis[0], tmp1, tmp2, xdim, ydim, zdim, gd0, gd0, gd3, 5, 5, 5, 1, 1, 1); /* Ittt */
      apply_separable_3d_filter(seq, basis[1], tmp1, tmp2, xdim, ydim, zdim, gd0, gd1, gd2, 5, 5, 5, 1, 1, 1); /* Iytt */
      apply_separable_3d_filter(seq, basis[2], tmp1, tmp2, xdim, ydim, zdim, gd0, gd2, gd1, 5, 5, 5, 1, 1, 1); /* Iyyt */
      apply_separable_3d_filter(seq, basis[3], tmp1, tmp2, xdim, ydim, zdim, gd0, gd3, gd0, 5, 5, 5, 1, 1, 1); /* Iyyy */
      apply_separable_3d_filter(seq, basis[4], tmp1, tmp2, xdim, ydim, zdim, gd1, gd0, gd2, 5, 5, 5, 1, 1, 1); /* Ixtt */
      apply_separable_3d_filter(seq, basis[5], tmp1, tmp2, xdim, ydim, zdim, gd1, gd1, gd1, 5, 5, 5, 1, 1, 1); /* Ixyt */
      apply_separable_3d_filter(seq, basis[6], tmp1, tmp2, xdim, ydim, zdim, gd1, gd2, gd0, 5, 5, 5, 1, 1, 1); /* Ixyy */
      apply_separable_3d_filter(seq, basis[7], tmp1, tmp2, xdim, ydim, zdim, gd2, gd0, gd1, 5, 5, 5, 1, 1, 1); /* Ixxt */
      apply_separable_3d_filter(seq, basis[8], tmp1, tmp2, xdim, ydim, zdim, gd2, gd1, gd0, 5, 5, 5, 1, 1, 1); /* Ixxy */
      apply_separable_3d_filter(seq, basis[9], tmp1, tmp2, xdim, ydim, zdim, gd3, gd0, gd0, 5, 5, 5, 1, 1, 1); /* Ixxx */
    break;
  default: printf("generate_3d_basis: basis_type %d not supported.\n"); exit(0); break;
  }
}

void energy_3d_basis(float ****basis, float ***energy, float ***tmp1, float ***tmp2, float ***tmp3,
		     int xdim, int ydim, int zdim, int basis_size)
{
  int b;

  printf("Basis energy:%dx%dx%d\n", xdim, ydim, zdim);
  fill_sequence(tmp2, 0.0, xdim, ydim, zdim);

  for (b=0; b<basis_size; b++) {
    square_sequence(basis[b], tmp1, xdim, ydim, zdim);
    sum_sequence(tmp1, tmp2, tmp2, xdim, ydim, zdim);
  }
  blur_3d(tmp2, energy, tmp1, tmp3, xdim, ydim, zdim);
}


void steer_3d_basis(float ****basis, float ***seq, float spatial_orient, float temporal_orient, 
		    int xdim, int ydim, int nframe, int basis_type)
{
  float cso = cos(spatial_orient), sso = sin(spatial_orient);
  float cto = cos(temporal_orient), sto = sin(temporal_orient);
  float cso2 = cso*cso, sso2 = sso*sso, cso3 = cso2*cso, sso3 = sso2*sso;
  float cto2 = cto*cto, sto2 = sto*sto, cto3 = cto2*cto, sto3 = sto2*sto;
  int x, y, t;

  printf("Steering: %dx%dx%d\n", xdim, ydim, nframe);
  switch ( basis_type ) {
  case G1_BASIS:
    for (t=0; t<nframe; t++) 
      for (y=0; y<ydim; y++) 
	for (x=0; x<xdim; x++) 
	  seq[t][y][x] =  basis[0][t][y][x]*cto + basis[1][t][y][x]*sso*sto + basis[2][t][y][x]*cso*sto;
    break;
  case G2_BASIS:
    for (t=0; t<nframe; t++) 
      for (y=0; y<ydim; y++) 
	for (x=0; x<xdim; x++) 
	  seq[t][y][x] = .5*(basis[0][t][y][x]*cto2 +  basis[5][t][y][x]*cso2*sto2 + basis[2][t][y][x]*sso2*sto2)
	    + basis[3][t][y][x]*cso*cto*sto + basis[1][t][y][x]*cto*sso*sto + basis[4][t][y][x]*cso*sso*sto2;
    break;
  case G3_BASIS:
    for (t=0; t<nframe; t++) 
      for (y=0; y<ydim; y++) 
	for (x=0; x<xdim; x++)
	  seq[t][y][x] = (basis[0][t][y][x]*cto3 + basis[3][t][y][x]*sso3*sto3 + basis[9][t][y][x]*cso3*sto3)/6.0 + 
	    .5*(basis[1][t][y][x]*cto2*sso*sto + basis[2][t][y][x]*cto*sso2*sto2 +
		basis[4][t][y][x]*cso*cto2*sto + basis[6][t][y][x]*cso*sso2*sto3 +
		basis[7][t][y][x]*cso2*cto*sto2 + basis[8][t][y][x]*cso2*sso*sto3) + 
	      basis[5][t][y][x]*cso*cto*sso*sto2;
    break;
  default: printf("steer_3d_basis: basis_type %d not supported.\n"); exit(0); break;
  }
}

float steer_3d_basis_point(float *basis, float spatial_orientation, float temporal_orientation, int basis_type)
{
  float cso = cos(spatial_orientation), sso = sin(spatial_orientation);
  float cto = cos(temporal_orientation), sto = sin(temporal_orientation);
  float cso2 = cso*cso, sso2 = sso*sso, cso3 = cso2*cso, sso3 = sso2*sso;
  float cto2 = cto*cto, sto2 = sto*sto, cto3 = cto2*cto, sto3 = sto2*sto;

  switch ( basis_type ) {
  case G1_BASIS:
    return(basis[0]*cto + basis[1]*sso*sto + basis[2]*cso*sto);
  case G2_BASIS:
    return(.5*(basis[0]*cto2 +  basis[5]*cso2*sto2 + basis[2]*sso2*sto2)
	    + basis[3]*cso*cto*sto + basis[1]*cto*sso*sto + basis[4]*cso*sso*sto2);
  case G3_BASIS:
    return((basis[0]*cto3 + basis[3]*sso3*sto3 + basis[9]*cso3*sto3)/6.0 + 
	    .5*(basis[1]*cto2*sso*sto + basis[2]*cto*sso2*sto2 +
		basis[4]*cso*cto2*sto + basis[6]*cso*sso2*sto3 +
		basis[7]*cso2*cto*sto2 + basis[8]*cso2*sso*sto3) + 
	   basis[5]*cso*cto*sso*sto2);
  default: printf("steer_3d_basis_point: basis_type %d not supported.\n"); exit(0); break;
  }
  return(0.0);
}

void read_descriptor(char* desc, int* xdim, int* ydim, int* nframe,
		     int *bytes_pixel)
{
  char descriptor[STRING], line[STRING], label[STRING], junk[STRING], junk2[STRING];
  FILE* d;
  int sets;

  sprintf(descriptor, "%s/descriptor", desc);
  if ((d = fopen(descriptor, "r")) == NULL) {
    printf("Cannot open descriptor input file %s\n", descriptor);
    exit(1);
  }
  
  *nframe = 0;
  *bytes_pixel = 1;  
  sets = 0;
  while (fgets(line, STRING, d) != NULL) {
    sscanf(line,"%s",label);
    if (!strcmp (label, "(_data_files"))
      sscanf(line,"%s%d%s", junk, nframe, junk2);
    else  if (!strcmp (label, "(_dimensions"))
      sscanf(line,"%s%d%d%s", junk, ydim, xdim, junk2);
    else  if (!strcmp (label, "(_data_sets"))
      sscanf(line,"%s%d%s", junk, &sets, junk2);
    else  if (!strcmp (label, "(_data_type")) {
      sscanf(line,"%s%s", junk, junk2);
      if ( junk2[0] == 'f')
	*bytes_pixel = 4;
      else
	*bytes_pixel = 1;
    }
  }
  printf("Reading %s: %dx%d %d frames  %d bytes/pixel\n", desc, *xdim, *ydim, *nframe, *bytes_pixel);
  fclose (d);
}

void write_picture(char* name, void** pic, int xdim, int ydim, int bytes_pixel)
{
  char file[STRING], command[STRING];
  FILE* o;
  struct stat stbuf;
  int y;

  if (stat(name, &stbuf) == -1)    {
    sprintf(command,"mkdir %s", name);
    system(command);
  }

  sprintf(file, "%s/data", name);
  if ((o = fopen(file, "wb")) == NULL) {
    printf("Cannot open output file %s\n", file);
    exit(1);
  }
  for (y=0; y<ydim; y++) 
    fwrite(pic[y], bytes_pixel, xdim, o); 
  fclose(o);
}

void write_vector(char* name, void* pic, int xdim, int bytes_pixel)
{
  char file[STRING], command[STRING];
  FILE* o;
  struct stat stbuf;

  if (stat(name, &stbuf) == -1)    {
    sprintf(command,"mkdir %s", name);
    system(command);
  }

  sprintf(file, "%s/data", name);
  if ((o = fopen(file, "wb")) == NULL) {
    printf("Cannot open output file %s\n", file);
    exit(1);
  }
  fwrite(pic, bytes_pixel, xdim, o); 
  fclose(o);
}

void write_descriptor(char* name, int xdim, int ydim,
		      int nframe, int bytes_pixel)
{
  char dat_type[30], file[STRING], command[STRING];
  FILE* d;
  struct stat stbuf;

  printf("Writing %s: %dx%d %d frames  %d bytes/pixel\n", name, xdim, ydim, nframe, bytes_pixel);
  if (stat(name, &stbuf) == -1)    {
    sprintf(command,"mkdir %s", name);
    system(command);
  }
  sprintf(file, "%s/descriptor", name);

  if (bytes_pixel == 1)
    strcpy(dat_type, "unsigned_1 ");
  else
    strcpy(dat_type, "float ");

  if ((d = fopen(file, "w")) == NULL) { 
    printf("Cannot open descriptor output file %s\n", file);
    exit(1);
  }
  if ( nframe > 1 ) 
    fprintf(d, "(_data_files %d)\n(_channels 1)\n(_dimensions %d %d)\n(_data_type %s)", nframe, ydim, xdim, dat_type);
  else 
    fprintf(d, "(_data_sets %d)\n(_channels 1)\n(_dimensions %d %d)\n(_data_type %s)", nframe, ydim, xdim, dat_type);

  fclose(d);
}

void write_sequence_frame(char* name, void **pic, 
			  int frame, int xdim, int ydim, int nframe,
			  int bytes_pixel)
{
  char file[STRING];
  FILE* o;
  int y;

  if ( nframe == 1 )
    sprintf(file, "%s/data", name);
  else 
    sprintf(file, "%s/data%d", name, frame);

  if ( (o = fopen(file, "wb") ) == NULL ) {
    printf("Couldn't write to %s\n", file);
    exit(0);
  }
  for (y=0; y<ydim; y++) 
    fwrite(pic[y], bytes_pixel, xdim, o); 
  fclose(o);
}

void write_sequence_frame_vector(char* name, void* pic, int frame, int
				 xdim, int bytes_pixel)
{
  char file[STRING];
  FILE* o;


  sprintf(file, "%s/data%d", name, frame);
  if ( (o = fopen(file, "wb") ) == NULL ) {
    printf("Couldn't write to %s\n", file);
    exit(0);
  }
  fwrite(pic, bytes_pixel, xdim, o); 
  fclose(o);
}

void read_sequence_frame_vector(char* name, void *pic, 
				int frame, int xdim, int bytes_pixel)
{
  char file[STRING];
  FILE* f;
  struct stat stbuf;


  sprintf(file, "%s/data%d", name, frame);
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }    
  if ((f = fopen(file, "r")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  fread(pic, bytes_pixel, xdim, f); 
  fclose(f);
}


void read_sequence_frame(char* name, void **pic, int frame, int xdim,
			 int ydim, int zdim, int bytes_pixel)
{
  char file[STRING];
  FILE* f;
  struct stat stbuf;
  int y;

  sprintf(file, "%s/data%d", name, frame);
  if ( stat(file, &stbuf) == -1 ) {
    printf("Error: Couldn't access %s. \n", file);
    exit(1);
  }    
  if ((f = fopen(file, "r")) == NULL) {
    printf("Cannot open input file %s\n", file);
    exit(1);
  }
  for (y=0; y<ydim; y++) 
    fread(pic[y], bytes_pixel, xdim, f); 

  fclose(f);
}

void write_sequence(char *name, void ***seq, int xdim, int ydim, int
		    nframe, int bytes_pixel)
{
  char file[STRING], command[STRING];
  FILE* o;
  int frame, y;
  struct stat stbuf;
  
  if (stat(name, &stbuf) == -1)    {
    sprintf(command,"mkdir %s", name);
    system(command);
  }

  for (frame = 0; frame < nframe; frame++) {
    if ( nframe > 1 ) {
      printf(".");
      fflush(stdout);
    }
    sprintf(file, "%s/data%d", name, frame);
    if ((o = fopen(file, "wb")) == NULL) {
      printf("Cannot open output file %s\n", file);
      exit(1);
    }
    for (y=0; y<ydim; y++) 
      fwrite(seq[frame][y], bytes_pixel, xdim, o);
    fclose(o);
  }
  if ( nframe > 1 ) 
    printf("\n");
}


void dump_float_picture(char *name, float **pic, unsigned char **tmp, int xdim, int ydim)
{
  write_descriptor(name, xdim, ydim, 1, sizeof(char));
  convert_float_char(pic, tmp, xdim, ydim, SAVE_POS);
  write_picture(name, (void**)tmp, xdim, ydim, sizeof(char));
}

void dump_char_picture(char *name, unsigned char **pic, int xdim, int ydim)
{
  write_descriptor(name, xdim, ydim, 1, sizeof(char));
  write_picture(name, (void**)pic, xdim, ydim, sizeof(char));
}

void generate_2d_basis(float **image, float ***basis, float **tmp, int xdim, int ydim, int basis_type)
{
static float g2f1[G2_FILTSIZE] = { .0094, .1148, .3964, -.0601, -.9213, -.0601, .3964, .1148, .0094 };
static float g2f2[G2_FILTSIZE] = { .0008, .0176, .1660, .6383, 1.0, .6383, .1660, .0176, .0008 };
static float g2f3[G2_FILTSIZE] = { -.0028, -.0480, -.3020, -.5806, 0.0, .5806, .3020, .0480, .0028 };
static float h2f1[G2_FILTSIZE] = { -.0098, -.0618, .0998, .7551, 0.0, -.7551, -.0998, .0618, .0098 };
static float h2f2[G2_FILTSIZE] = { .0008, .0176, .1660, .6383, 1.0, .6383, .1660, .0176, .0008 };
static float h2f3[G2_FILTSIZE] = { -.0020, -.0354, -.2225, -.4277, 0.0, .4277, .2225, .0354, .0020 };
static float h2f4[G2_FILTSIZE] = { .0048, .0566, .1695, -.1889, -.7349, -.1889, .1695, .0566, .0048 };

  printf("Generating basis:%dx%d\n", xdim, ydim);
  switch ( basis_type ) {
  case G2_BASIS:
      apply_separable_2d_filter(image, basis[0], tmp,  xdim, ydim,  g2f1, g2f2, 9, 9,  1, 1); /* Itt */
      apply_separable_2d_filter(image, basis[1], tmp,  xdim, ydim,  g2f3, g2f3, 9, 9,  1, 1); /* Iyt */
      apply_separable_2d_filter(image, basis[2], tmp,  xdim, ydim,  g2f2, g2f1, 9, 9,  1, 1); /* Iyy */
      apply_separable_2d_filter(image, basis[3], tmp,  xdim, ydim,  h2f1, h2f2, 9, 9,  1, 1); /* Ixt */
      apply_separable_2d_filter(image, basis[4], tmp,  xdim, ydim,  h2f4, h2f3, 9, 9,  1, 1); /* Ixy */
      apply_separable_2d_filter(image, basis[5], tmp,  xdim, ydim,  h2f3, h2f4, 9, 9,  1, 1); /* Ixx */
      apply_separable_2d_filter(image, basis[6], tmp,  xdim, ydim,  h2f2, h2f1, 9, 9,  1, 1); /* Ixx */
    break;
  case G4_BASIS:
    break;
  default: printf("generate_2d_basis: basis_type %d not supported.\n"); exit(0); break;
  }
}

void steer_2d(float ***basis, float **o, int xdim, int ydim, double t)
{
  int ix, iy;
  float ct, st, ct2, st2, ct3, st3, ctst, ct2st, ctst2, g2, h2;
  float ga, gb, gc, ha, hb, hc, hd;

  ct = cos(t);  ct2 = ct*ct;  ct3 = ct2*ct;
  st = sin(t);  st2 = st*st;  st3 = st2*st;
  ctst = ct*st;  ct2st = ct2*st;  ctst2 = ct*st2;
  for (iy=0; iy <ydim; iy++) {
    for (ix=0; ix<xdim; ix++) {
      ga = basis[0][iy][ix];
      gb = basis[1][iy][ix];
      gc = basis[2][iy][ix];
      ha = basis[3][iy][ix];
      hb = basis[4][iy][ix];
      hc = basis[5][iy][ix];
      hd = basis[6][iy][ix];
      g2 = (ct2*ga - 2*gb*ctst + gc*st2);
      h2 = (3.0*(hb*ct2st - hc*ctst2) + hd*st3 - ct3*ha);
      o[iy][ix] = g2*g2 + h2*h2;
    }
  }
}


void downsample_n(float **in, float **tmp, float **tmp2, float **tmp3,
		  float **out, int ixdim, int iydim, int *oxdim, int
		  *oydim, int times)
{
  static float gauss[5]={0.0625, 0.25, 0.375, 0.25, 0.0625};

  switch ( times ) {
  case 0: 
    apply_separable_2d_filter(in, out, tmp, ixdim, iydim, gauss, gauss, 5, 5, 1, 1);
    *oxdim = ixdim;     *oydim = iydim; 
    break;
  case 1: 
    apply_separable_2d_filter(in, out, tmp, ixdim, iydim, gauss, gauss, 5, 5, 2, 2);
    *oxdim = ixdim/2;     *oydim = iydim/2; 
    break;
  case 2:
    apply_separable_2d_filter(in, tmp2, tmp, ixdim, iydim, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp2, out, tmp, ixdim/2, iydim/2, gauss, gauss, 5, 5, 2, 2);
    *oxdim = ixdim/4;     *oydim = iydim/4; 
    break;
  case 3:
    apply_separable_2d_filter(in, tmp2, tmp, ixdim, iydim, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp2, tmp3, tmp, ixdim/2, iydim/2, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp3, out, tmp2, ixdim/4, iydim/4, gauss, gauss, 5, 5, 2, 2);
    *oxdim = ixdim/8;     *oydim = iydim/8; 
    break;
  case 4:
    apply_separable_2d_filter(in, tmp2, tmp, ixdim, iydim, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp2, tmp3, tmp, ixdim/2, iydim/2, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp3, tmp, tmp2, ixdim/4, iydim/4, gauss, gauss, 5, 5, 2, 2);
    apply_separable_2d_filter(tmp, out, tmp2, ixdim/8, iydim/8, gauss, gauss, 5, 5, 2, 2);
    *oxdim = ixdim/16;     *oydim = iydim/16; 
    break;
  }
}

base_constructor* create_base_constructor(int xdim, int ydim)
{
  base_constructor *b = (base_constructor*)malloc(sizeof(base_constructor));
  b->xdim = xdim;    b->ydim = ydim;
  b->ixdim = xdim/pow2(DOWNSAMPLE_IMAGE);     b->iydim = ydim/pow2(DOWNSAMPLE_IMAGE);
  b->bxdim = b->ixdim;                           b->bydim = b->iydim*ORIENTATIONS;
   
  printf("input: %dx%d  downsampled %dx%d  output %dx%d\n",
	 xdim, ydim, b->ixdim, b->iydim,  b->bxdim, b->bydim);
  b->tmp = (float***)allocate_sequence(xdim, ydim, 3, sizeof(float));
  b->input_float = (float**)allocate_picture(xdim, ydim, sizeof(float));
  b->input_downsampled = (float**)allocate_picture(b->ixdim, b->iydim, sizeof(float));
  b->input_basis = (float***)allocate_sequence(b->ixdim, b->iydim, 7, sizeof(float));
  b->input_energy = (float***)allocate_sequence(b->ixdim, b->iydim, 7, sizeof(float));
  b->input_total_energy = (float**)allocate_picture(b->ixdim, b->iydim, sizeof(float));
  b->input_total_energy_blurred = (float**)allocate_picture(b->ixdim, b->iydim, sizeof(float));
  b->output_float = (float**)allocate_picture(b->bxdim, b->bydim, sizeof(float));
  b->SIGMA = .001;
  return(b);
}

void base_representation(base_constructor *b, void **input, int input_size, void **output, int output_size)
{
  static float gauss2[3]={ .25, .5, .25 };
  int x, y, s;
  unsigned char **ichar, **ochar;
  float **ifloat, **ofloat;

  if ( input_size == sizeof(char) ) {
    ichar = (unsigned char**)input;
    convert_char_float(ichar, b->input_float, b->xdim, b->ydim, 1);
    downsample_n(b->input_float, b->tmp[0], b->tmp[1], b->tmp[2], b->input_downsampled, b->xdim, b->ydim, 
		 &b->ixdim, &b->iydim, DOWNSAMPLE_IMAGE);
  } else {
    ifloat = (float**)input;
    downsample_n(ifloat, b->tmp[0], b->tmp[1], b->tmp[2], b->input_downsampled, b->xdim, b->ydim, 
		 &b->ixdim, &b->iydim, DOWNSAMPLE_IMAGE);
  }

  generate_2d_basis(b->input_downsampled, b->input_basis, b->tmp[0], b->ixdim, b->iydim, G2_BASIS);
  fill_image(b->input_total_energy, (float)0.0, b->ixdim, b->iydim);
  for (s=0; s<ORIENTATIONS; s++) {
    steer_2d(b->input_basis, b->input_energy[s], b->ixdim, b->iydim, 
	     (double)PI*(float)s/(float)ORIENTATIONS);
    add_image(b->input_energy[s], b->input_total_energy, b->input_total_energy, b->ixdim, b->iydim);
  }
  apply_separable_2d_filter(b->input_total_energy, b->input_total_energy_blurred, b->tmp[0],  
			    b->ixdim, b->iydim,  gauss2, gauss2, 3, 3, 1, 1);
  add_cimage(b->input_total_energy_blurred, b->SIGMA, b->input_total_energy_blurred, b->ixdim, b->iydim);
  for (s=0; s<ORIENTATIONS; s++) {
    divide_image(b->input_energy[s], b->input_total_energy_blurred, b->input_energy[s], b->ixdim, b->iydim);
    if ( output_size == sizeof(float) ) {
      ofloat = (float**)output;
      for (y=0; y<b->iydim; y++)
	for (x=0; x<b->ixdim; x++) 
	  ofloat[s*b->iydim+y][x] = b->input_energy[s][y][x];
    } else {
      for (y=0; y<b->iydim; y++)
	for (x=0; x<b->ixdim; x++)
	  b->output_float[s*b->iydim+y][x] = b->input_energy[s][y][x];
    }
  }

  if ( output_size == sizeof(char) ) {
    ochar = (unsigned char**)output;
    convert_float_char(b->output_float, ochar, b->bxdim, b->bydim, SAVE_POS);
  }
}
