#ifndef _TOOL_H
#define _TOOL_H

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>

#define MAX_CHARS 128

typedef struct base_constructor {
  float **params;
  float **input_float, **input_downsampled, ***input_basis, ***input_energy;
  float **input_total_energy, **input_total_energy_blurred, **output_float, **output_tmp, ***tmp;
  int xdim, ydim, zdim, ixdim, iydim,  bxdim, bydim;
  float SIGMA;
} base_constructor;

#define PI                3.14159

/*
#define DOWNSAMPLE_IMAGE  3
*/
#define DOWNSAMPLE_IMAGE  2

#define EPS               1.0e-9
#define ORIENTATIONS      6

base_constructor* create_base_constructor(int xdim, int ydim);
void base_representation(base_constructor *b, void **input, int input_size, void **output, int output_size);


#define MAXFRAMES  600
#define STRING     256


#define SAVE_POSNEG  0
#define SAVE_POS     1

#define G1_BASIS 1
#define G2_BASIS 2
#define G3_BASIS 3
#define G4_BASIS 4

#define G1_BASIS_SIZE  3
#define G2_BASIS_SIZE  6
#define G3_BASIS_SIZE  10
#define G4_BASIS_SIZE  15

#define G1_2D_BASIS_SIZE             1
#define G1_3D_BASIS_SIZE             3
#define G2_3D_BASIS_SIZE             6
#define G3_3D_BASIS_SIZE             10

#define G2_2D_BASIS_SIZE             3
#define H2_2D_BASIS_SIZE             4
#define G2_H2_2D_BASIS_SIZE          7

#define G0_FILTSIZE                     5 
#define G1_FILTSIZE                     5 
#define G2_FILTSIZE                     9 

void **allocate_picture(int xdim, int ydim, int size);
void fill_image(float **res, float val, int xdim, int ydim);
void add_image(float **im1, float **im2, float **res, int xdim, int ydim);
void subtract_image(float **im1, float **im2, float **res, int xdim, int ydim);
void add_cimage(float **im1, float c, float **res, int xdim, int ydim);
void divide_image(float **im1, float **im2, float **res, int xdim, int ydim);
/* 
void dump_float_picture(unsigned char *name, float **pic, unsigned char **tmp, int xdim, int ydim);
void dump_char_picture(unsigned char *name, unsigned char **pic, int xdim, int ydim);
*/
float energy_image(float **image, int xdim, int ydim);
float energy_vector(float *vec, int dim);
float euclidean_distance_vector(float *v1, float *v2, int dim);
float weighted_euclidean_distance_vector(float *v1, float *v2, float *w, int dim);
float mahalanobis_distance_vector(float *v1, float *v2, float *eigenval, int dim);
int random_integer(int s, int e);
float random_float(float f0, float f1);
int pow2(int n);
void draw_line(unsigned char **pic, float p0[2], float p1[2], int xdim, int ydim, unsigned char val);
void read_vector(char* name, void* pic, int xdim, int bytes_pixel);
void read_picture(char* name, void **pic, int xdim, int ydim,
		  int bytes_pixel);
void write_descriptor(char* name, int xdim, int ydim, int
		      nframe, int bytes_pixel);
void write_picture(char* name, void **pic, int xdim, int
		   ydim, int bytes_pixel);
void write_sequence_frame(char* name, void **pic, 
			  int frame, int xdim, int ydim, int nframe,
			  int bytes_pixel);
void read_sequence(char* name, void ***seq, int xdim, int ydim, int
		   nframe, int bytes_pixel);
void read_sequence_frame_vector(char* name, void *pic, 
				int frame, int xdim, int bytes_pixel);
void read_descriptor(char* desc, int* xdim, int* ydim, int* nframe,
		     int* bytes_pixel);
void read_sequence_frame(char* name, void **pic, int frame, int xdim,
			 int ydim, int nframe, int bytes_pixel);
void write_vector(char* name, void* pic, int xdim, int bytes_pixel);
void write_sequence_frame_vector(char* name, void* pic, int frame, int
				 xdim, int bytes_pixel);

#endif
